package ava.parser

import cats.effect.IO
import cats.effect.Resource
import cats.effect.unsafe.IORuntime
import java.io.ByteArrayInputStream
import java.io.InputStream

class TokenizerTests extends munit.FunSuite:
  implicit val runtime: IORuntime = cats.effect.unsafe.IORuntime.global

  private val EmptyStream: InputStream = new ByteArrayInputStream(Array[Byte]())

  test("should produce the EOF token for an empty stream") {
    run(newCR(EmptyStream)) { tokenizer =>
      IO(assertEquals(tokenizer.next(), Right(Token.Eof)))
    }
  }

  test(
    "should produce the EOF token for a stream of only whitespace characters"
  ) {
    run(newCR(stringStream(" \t\n\r\n   "))) { tokenizer =>
      IO(assertEquals(tokenizer.next(), Right(Token.Eof)))
    }
  }

  test(
    "should produce an error, followed by EOF, if only a backslash chararacter is present"
  ) {
    assertTokens(
      " \\ ",
      Left(Tokenizer.Error.BackSlashNotAllowed(pos(2, 1, 2)))
    )
  }

  test("should capture all fixed, single-character tokens") {
    assertTokens(
      "(),:.#",
      Right(Token.OpenParen),
      Right(Token.CloseParen),
      Right(Token.Comma),
      Right(Token.Colon),
      Right(Token.Dot),
      Right(Token.Tuple)
    )
  }

  test("should not recognize comments if preceded by any non-whitespace") {
    assertTokens(
      "( --",
      Right(Token.OpenParen),
      Right(Token.Generic("--"))
    )
  }

  test("should recognize potential comments that are just a dash (space)") {
    assertTokens("- ", Right(Token.Generic("-")))
  }

  test("should recognize potential comments that are just a dash (tab)") {
    assertTokens("-\t", Right(Token.Generic("-")))
  }

  test("should recognize potential comments that are just a dash (newline)") {
    assertTokens("-\n", Right(Token.Generic("-")))
  }

  test("should handle basic comments") {
    assertTokens("-- Comment", Right(Token.Comment("Comment")))
  }

  test("should handle comments, if only whitespace precedes them") {
    assertTokens("\t   -- Comment", Right(Token.Comment("Comment")))
  }

  test("should handle successive lines of comments") {
    assertTokens(
      "--c1\n--c2\n--c3",
      Right(Token.Comment("c1")),
      Right(Token.Comment("c2")),
      Right(Token.Comment("c3"))
    )
  }

  test("should ignore comment contents") {
    assertTokens("-- a\\12 3\t4", Right(Token.Comment("a\\12 3\t4")))
  }

  test("should throw an error if EOF is reached inside a string literal") {
    assertTokens(
      "\"",
      Left(
        Tokenizer.Error.PrematureEof(
          List(Tokenizer.State.InDoubleQuote(pos(1, 1, 1)))
        )
      )
    )
  }

  test("should throw an error if EOF is reached inside a character literal") {
    assertTokens(
      "\'",
      Left(
        Tokenizer.Error.PrematureEof(
          List(Tokenizer.State.InSingleQuote(pos(1, 1, 1)))
        )
      )
    )
  }

  test("should capture all supported char escapes in a string literal") {
    assertTokens(
      "\"\\n\\r\\b\\f\\t\\\\\\\"\\\'\"",
      Right(Token.StringLiteral("\n\r\b\f\t\\\"\'", Nil))
    )
  }

  test("should capture all supported char escapes in character literals") {
    assertTokens("\'\\n\'", Right(Token.CharacterLiteral('\n', Nil)))
    assertTokens("\'\\r\'", Right(Token.CharacterLiteral('\r', Nil)))
    assertTokens("\'\\b\'", Right(Token.CharacterLiteral('\b', Nil)))
    assertTokens("\'\\f\'", Right(Token.CharacterLiteral('\f', Nil)))
    assertTokens("\'\\t\'", Right(Token.CharacterLiteral('\t', Nil)))
    assertTokens("\'\\\\\'", Right(Token.CharacterLiteral('\\', Nil)))
    assertTokens("\'\\\"\'", Right(Token.CharacterLiteral('"', Nil)))
    assertTokens("\'\\'\'", Right(Token.CharacterLiteral('\'', Nil)))
  }

  test("should handle a basic string literal") {
    assertTokens("\"foo\"", Right(Token.StringLiteral("foo", Nil)))
  }

  test("should handle a string literal with an invalid character escape") {
    // Note: The character escape is simply not emitted, but we are able to
    // successfully close out the string, which allows us to recover and parse
    // more content.
    assertTokens(
      "\"foo\\!bar\"",
      Right(
        Token.StringLiteral(
          "foobar",
          List(Tokenizer.Error.InvalidCharEscape(pos(6, 1, 6), '!'))
        )
      )
    )
  }

  private def assertTokens(
    source: String,
    expectedOutput: Either[Tokenizer.Error, Token]*
  ): Unit =
    run(newCR(stringStream(source))) { tokenizer =>
      consumeAll(tokenizer).map { tokens =>
        assertEquals(tokens, expectedOutput.toList)
      }
    }

  private def consumeAll(tokenizer: Tokenizer)
    : IO[List[Either[Tokenizer.Error, Token]]] =
    fs2.Stream
      .repeatEval(IO(tokenizer.next()))
      .takeWhile(_ != Right(Token.Eof))
      .compile
      .toList

  private def pos(
    charInFile: Long,
    lineNumber: Long,
    charInLine: Long
  ): SourcePosition =
    SourcePosition(charInFile, lineNumber, charInLine)

  private def stringStream(contents: String): InputStream =
    new ByteArrayInputStream(contents.getBytes())

  private def newCR(is: InputStream): CharacterReader =
    CharacterReader.forInputStream(is)

  private def run(reader: CharacterReader)(testF: Tokenizer => IO[Unit]): Unit =
    Resource
      .make(
        acquire = IO(new Tokenizer(reader))
      )(release = t => IO(t.close()))
      .use(testF)
      .unsafeRunSync()
